from simio_lisa.simio_tables import *
import logging
import pandas as pd
import os
import plotly.express as px
from plotly.offline import plot
import time
from abc import ABC, abstractmethod


class SimioPlotter(ABC):
    def __init__(self,
                 output_tables,
                 logger_level: int = logging.INFO,
                 **kwargs):
        """
        Parent class.
        :param output_tables: DICT containing all tables
        :param **x_axis/y_axis/time_axis: column to be used as x/y/time axis
        :param **legend_col: column to use to distinguish colors/legend
        :param **objects_dict: dictionary to distinguish the groups of entities to be compared together
        """
        self._tables_names = None
        self._tables = output_tables
        # Instance Tables
        self._x_axis = kwargs.get('x_axis', None)
        self._y_axis = kwargs.get('y_axis', None)
        self._time_axis = kwargs.get('time_axis', None)
        self._objects_dict = kwargs.get('objects_dict', None)
        self._legend_col = kwargs.get('legend_col', None)

        logging.getLogger().setLevel(logger_level)

    @abstractmethod
    def plot(self, tables, kind):
        """
        Force all subclasses to have a plot method
        """
        pass

    @property
    def tables(self):
        return self._tables

    @property
    def tables_names(self):
        return self._tables_names

    @property
    def time_axis(self):
        return self._time_axis

    @time_axis.setter
    def time_axis(self, new_value):
        self._time_axis = new_value

    @property
    def y_axis(self):
        return self._y_axis

    @y_axis.setter
    def y_axis(self, new_value):
        self._y_axis = new_value

    @property
    def x_axis(self):
        return self._x_axis

    @x_axis.setter
    def x_axis(self, new_value):
        self._x_axis = new_value

    @property
    def objects_dict(self):
        return self._objects_dict

    @objects_dict.setter
    def objects_dict(self, new_value):
        self._objects_dict = new_value

    @property
    def legend_col(self):
        return self._legend_col

    @legend_col.setter
    def legend_col(self, new_value):
        self._legend_col = new_value


class SimioTimeSeries(SimioPlotter):
    def __init__(self,
                 output_tables,
                 logger_level: int = logging.INFO,
                 **kwargs):
        """
        Class child of SimioPlotter to plot time series. Necessary in kwargs: time_axis and y_axis.
        When using plot_columns (plot_tables) y_axis (tables) can be a list.
        """
        SimioPlotter.__init__(self,
                              output_tables,
                              logger_level,
                              **kwargs)

    def plot_columns(self, table: str):
        """
        Plot TimeSeries comparing different columns (y_Axis should be a list of columns,
        only one table should be provided)
        """
        input_data = self.tables[table]
        time_axis = self.time_axis
        y_axis = self.y_axis

        input_data[time_axis] = pd.to_datetime(input_data[time_axis])
        fig = px.line(input_data, x=time_axis, y=y_axis)
        return fig

    def plot_tables(self, tables):
        """
        Plot TimeSeries comparing different tables (time_axis and y_axis should have the same name in all the tables)
        :param tables:
        :return:
        """
        if type(tables) is str:
            input_data = self.tables[tables]
            input_data['source'] = tables
        else:
            input_data = pd.DataFrame()
            for t in tables:
                aux = self.tables[t]
                aux['source'] = t
                input_data = input_data.append(aux, ignore_index=True)
        time_axis = self.time_axis
        y_axis = self.y_axis

        input_data[time_axis] = pd.to_datetime(input_data[time_axis])

        fig = px.line(input_data, x=time_axis, y=y_axis, color='source')
        return fig

    def plot(self, tables, kind):
        if kind == 'time_series_columns':
            fig = self.plot_columns(table=tables)
            plot(fig)
        elif kind == 'time_series_tables':
            fig = self.plot_tables(tables=tables)
            plot(fig)
        else:
            raise ValueError(f'Kind {kind} not defined')


class SimioBarPie(SimioPlotter):
    def __init__(self,
                 output_tables,
                 logger_level: int = logging.INFO,
                 **kwargs):
        """
        Class child of SimioPlotter to plot bar plots or pie charts. Necessary in kwargs: x_axis, y_axis and
        objects_dict (time_axis necessary only for method plot_bars_time_series)
        :param objects_dict: dictionary grouping the objects to compare
        """
        SimioPlotter.__init__(self,
                              output_tables,
                              logger_level,
                              **kwargs)

    def plot_bars(self, tables: str):

        if type(tables) is str:
            input_data = self.tables[tables]
            input_data['source'] = tables
        else:
            input_data = pd.DataFrame()
            for t in tables:
                aux = self.tables[t]
                aux['source'] = t
                input_data = input_data.append(aux, ignore_index=True)
        y_axis = self.y_axis
        x_axis = self.x_axis
        object_groups_dict = self.objects_dict
        input_data[y_axis] = input_data[y_axis].astype(float)  # otherwise the column is lost when grouping

        f = {}
        for k in object_groups_dict.keys():
            input_data_plt = input_data[input_data[x_axis].isin(object_groups_dict[k])].copy(deep=True)
            input_data_plt = input_data_plt.groupby(by=x_axis, as_index=False).mean()
            fig = px.bar(input_data_plt, x=x_axis, y=y_axis, barmode="group")
            f[k] = fig
        return f

    def plot_pie(self, tables: str):

        if type(tables) is str:
            input_data = self.tables[tables]
            input_data['source'] = tables
        else:
            input_data = pd.DataFrame()
            for t in tables:
                aux = self.tables[t]
                aux['source'] = t
                input_data = input_data.append(aux, ignore_index=True)
        y_axis = self.y_axis
        x_axis = self.x_axis
        object_groups_dict = self.objects_dict
        input_data[y_axis] = input_data[y_axis].astype(float)  # otherwise the column is lost when grouping

        f = {}
        for k in object_groups_dict.keys():
            input_data_plt = input_data[input_data[x_axis].isin(object_groups_dict[k])].copy(deep=True)
            input_data_plt = input_data_plt.groupby(by=x_axis, as_index=False).mean()
            fig = px.pie(input_data_plt, values=y_axis, names=x_axis,
                         title=x_axis + ' utilization quantified in terms of ' + y_axis + ', group ' + k)
            f[k] = fig
        return f

    def plot_bars_time_series(self, tables: str):
        if type(tables) is str:
            input_data = self.tables[tables]
            input_data['source'] = tables
        else:
            input_data = pd.DataFrame()
            for t in tables:
                aux = self.tables[t]
                aux['source'] = t
                input_data = input_data.append(aux, ignore_index=True)
        y_axis = self.y_axis
        x_axis = self.x_axis
        time_axis = self.time_axis
        object_groups_dict = self.objects_dict
        input_data[time_axis] = input_data[time_axis].astype(str)
        input_data[y_axis] = input_data[y_axis].astype(float)  # otherwise the column is lost when grouping

        f = {}
        for k in object_groups_dict.keys():
            input_data_plt = input_data[input_data[x_axis].isin(object_groups_dict[k])].copy(deep=True)
            fig = px.bar(input_data_plt, x=x_axis, y=y_axis, barmode="group", facet_col=time_axis)
            f[k] = fig
        return f

    def plot(self, tables, kind):
        if kind == 'bars_plot':
            fig = self.plot_bars(tables=tables)
            for k in fig.keys():
                plot(fig[k])
                time.sleep(.5)
        elif kind == 'pie_plot':
            fig = self.plot_pie(tables=tables)
            for k in fig.keys():
                plot(fig[k])
                time.sleep(.5)
        elif kind == 'bars_time_series_plot':
            fig = self.plot_bars_time_series(tables=tables)
            for k in fig.keys():
                plot(fig[k])
                time.sleep(.5)
        else:
            raise ValueError(f'Kind {kind} not defined')


class SimioBox(SimioPlotter):
    def __init__(self,
                 output_tables,
                 logger_level: int = logging.INFO,
                 **kwargs):
        """

        :param project_path:
        :param project_name:
        :param project_path:
        :param model_name:
        """
        SimioPlotter.__init__(self,
                              output_tables,
                              logger_level,
                              **kwargs)

    def plot_box(self, tables):
        """
        Class child of SimioPlotter to plot box plots. Necessary in kwargs: x_axis and y_axis.
        """
        if type(tables) is str:
            input_data = self.tables[tables]
            input_data['source'] = tables
        else:
            input_data = pd.DataFrame()
            for t in tables:
                aux = self.tables[t]
                aux['source'] = t
                input_data = input_data.append(aux, ignore_index=True)
        x_axis = self.x_axis
        y_axis = self.y_axis
        fig = px.box(input_data, x=x_axis, y=y_axis, color=x_axis, notched=True)
        return fig

    def plot(self, tables, kind):
        if kind == 'box_plot':
            fig = self.plot_box(tables=tables)
            plot(fig)
        else:
            raise ValueError(f'Kind {kind} not defined')


class SimioStackedBars(SimioPlotter):
    def __init__(self,
                 output_tables,
                 logger_level: int = logging.INFO,
                 **kwargs):
        """
        Class child of SimioPlotter to plot stacked bars. Necessary in kwargs: x_axis, y_axis and legend_col.
        """
        SimioPlotter.__init__(self,
                              output_tables,
                              logger_level,
                              **kwargs)

    def plot_stacked_bars(self, tables):
        if type(tables) is str:
            input_data = self.tables[tables]
            input_data['source'] = tables
        else:
            input_data = pd.DataFrame()
            for t in tables:
                aux = self.tables[t]
                aux['source'] = t
                input_data = input_data.append(aux, ignore_index=True)
        x_axis = self.x_axis
        y_axis = self.y_axis
        legend_col = self.legend_col
        fig = px.bar(input_data, x=x_axis, y=y_axis, color=legend_col)
        return fig

    def plot(self, tables, kind):
        if kind == 'stacked_bars':
            fig = self.plot_stacked_bars(tables=tables)
            plot(fig)
        else:
            raise ValueError(f'Kind {kind} not defined')


def create_object_processing_table(tables):
    """
    Create a table with all the processes 'operation' happening to each entity 'object_id' and their durations extracted
     from 'start' and 'stop' columns.
    The processes come from two different tables defined in "operation1""table" and "operation2""table"
    :param tables: DICT
    :return: tables DICT (with the new table added with name "ObjectProcessingTable"
    """
    obj_processing_metadata = {'operation1': {'table': 'OutputLoadingUnloadingTimes',
                                              'columns':
                                                  {'duration': {'start': 'OperationBegins',
                                                                'stop': 'OperationFinish'},
                                                   'operation': 'OperationType',
                                                   'object_id': 'CubeName'}
                                              },
                               'operation2': {'table': 'OutputPalletProcessing',
                                              'columns':
                                                  {'duration': {'start': 'ProcessingStartTime',
                                                                'stop': 'ProcessingEndTime'},
                                                   'operation': 'RetortId',
                                                   'object_id': 'MaterialName'}
                                              }}

    table_obj_processing = pd.DataFrame()
    for o in obj_processing_metadata.keys():
        metadata = obj_processing_metadata[o]
        table_aux = tables.output_tables[metadata['table']]
        table_aux['duration'] = (table_aux[metadata['columns']['duration']['stop']] - table_aux[
            metadata['columns']['duration']['start']]).dt.total_seconds()
        table_aux = table_aux[['duration', metadata['columns']['operation'], metadata['columns']['object_id']]]
        table_aux.rename(
            columns={metadata['columns']['operation']: 'operation', metadata['columns']['object_id']: 'object_id'},
            inplace=True)
        table_obj_processing = table_obj_processing.append(table_aux, ignore_index=True)
    tables.output_tables['ObjectProcessingTable'] = table_obj_processing
    return tables


if __name__ == '__main__':
    path_output_tables = os.path.normpath(os.environ['PATHOUTPUTTABLES'])
    project_path = os.path.normpath(os.environ['PROJECTPATH'])
    project_name = os.environ['PROJECTNAME']
    model_name = os.environ['MODELNAME']

    simio_tables_ = SimioTables(path_to_project=project_path,
                                model_file_name=project_name,
                                model_name=model_name)
    simio_tables_.load_output_tables()
    print('Create Object Processing Table')
    output_tables_new_ = create_object_processing_table(simio_tables_)
    x_axis_ = 'object_id'
    y_axis_ = 'duration'
    operations_id = 'operation'
    simio_object_processing_plotter = SimioStackedBars(
        output_tables=output_tables_new_.output_tables,
        logger_level=logging.INFO,
        x_axis=x_axis_,
        y_axis=y_axis_,
        legend_col=operations_id)
    simio_object_processing_plotter.plot(tables='ObjectProcessingTable', kind='stacked_bars')



import os
import pandas as pd
from simio_lisa.utils import *
from simio_lisa.load_simio_project import get_model_metadata, get_model_tables_path, get_output_table_names, \
    get_input_table_names
from abc import ABC, abstractmethod


class SimioTable(ABC):
    def __init__(self, model_metadata, table_path: str):
        self.__model_metadata = model_metadata
        self.__table_path = table_path
        self.__table_schema = self.load_schema()
        self.__table_ddl = self.create_table_ddl()

    def create_table_ddl(self):
        table_name = self.table_path.split(os.sep)[-1].split('.')[0]
        ddl_table = f'CREATE TABLE {table_name} ('
        for col, data_type in self.table_schema.items():
            ddl_table += f'\n\t{col} {sql_data_types[data_type]},'
        ddl_table += '\b\n);'
        return ddl_table

    def load_schema(self):
        """
        Load data types of columns.
        :return: Dictionary mapping column name -> data type
        """
        list_tables = extract_list_from_dom(dom_object=self.model_metadata, tag_name='Table')
        attribute_value = self.table_path.split(os.path.sep)[-1].split('.')[0]
        dom_tables = filter_dom_by_attribute(list_of_doms=list_tables,
                                             attribute_name='Name',
                                             attribute_value=attribute_value)[0]
        metadata_results = dict()
        for simio_data_type, python_data_type in simio_data_types.items():
            column_names = extract_list_from_dom(dom_object=dom_tables,
                                                 tag_name=simio_data_type,
                                                 attribute_name='Name')
            for col in column_names:
                metadata_results[col] = python_data_type
        return metadata_results

    @abstractmethod
    def get_single_row_value(self, dom_row_list: list):
        """
        Get value of a single scenario's response. This method must be redefined in each of the child classes.
        :param dom_row_list: Single row provided as a list of one term.
        :return: Value of such observation.
        """
        pass

    def add_row_to_table(self,
                         result_dict: dict,
                         dom_row: minidom.Document,
                         column_names: set,
                         tag_name,
                         attribute_name):
        """
        Load row to table
        :param result_dict: Dictionary of results to be updated.
        :param dom_row: List of.
        :param column_names: Set of response names to iterate over.
        :param tag_name: tag of the .xml table to look for column names.
        :param attribute_name: attribute in tag_name with the column names.
        :return: Nothing. Result dictionary is updated.
        """
        row = dom_row.getElementsByTagName(tag_name)
        for resp in column_names:
            response = filter_dom_by_attribute(list_of_doms=row,
                                               attribute_name=attribute_name,
                                               attribute_value=resp)
            value_response = self.get_single_row_value(dom_row_list=response)
            result_dict[resp].append(value_response)

    @staticmethod
    def get_column_names(row_doms, tag_name, attribute_name):
        """

        :param row_doms: list of dom of rows.
        :param tag_name: tag of the .xml table to look for column names.
        :param attribute_name: attribute in tag_name with the column names.
        :return: list with the column names.
        """
        column_names = list()
        for i in row_doms:
            column_names.extend(extract_list_from_dom(dom_object=i,
                                                      tag_name=tag_name,
                                                      attribute_name=attribute_name))
        column_names = set(column_names)
        return column_names

    def load_single_table(self, tag_name='StateValue', attribute_name='Name') -> Union[pd.DataFrame, None]:
        """
        Load results of single experiment
        :param tag_name: tag of the .xml table to look for column names. The default is for the output tables.
        :param attribute_name: attribute in tag_name with the column names. The default is for the output tables.
        :return: DataFrame with result of experiment.
        """
        try:
            project_xml = minidom.parse(self.table_path)
        except FileNotFoundError:
            warnings.warn(f'Table {self.table_path} has not been ran yet.')
            return None
        row_doms = extract_list_from_dom(dom_object=project_xml,
                                         tag_name='Row')
        if len(row_doms) == 0:
            warnings.warn(f'Table {self.table_path} is empty')
            return None

        column_names = self.get_column_names(row_doms=row_doms, tag_name=tag_name, attribute_name=attribute_name)
        results_dict = {col: [] for col in column_names}
        for row in row_doms:
            self.add_row_to_table(result_dict=results_dict,
                                  dom_row=row,
                                  column_names=column_names,
                                  tag_name=tag_name,
                                  attribute_name=attribute_name)
        results_df = pd.DataFrame(results_dict)
        results_df = self.apply_data_type(results_df)
        return results_df

    def apply_data_type(self, results_df):
        """
        Force data type in each column using the schema simio_data_types in utils.
        :param results_df: df with the tables extracted.
        :return: initial df with the type of each column applied.
        """
        for col, val in self.table_schema.items():
            if val in [int, float, bool]:
                results_df[col] = results_df[col].astype(val)
            elif val == datetime.datetime:
                results_df[col] = pd.to_datetime(results_df[col])
            elif val == str:
                pass
            else:
                raise NotImplementedError(f'{val} is not a recognized data type.')

        return results_df

    # Getters and setters
    @property
    def model_metadata(self):
        return self.__model_metadata

    @model_metadata.setter
    def model_metadata(self, model_metadata):
        self.__model_metadata = model_metadata

    @property
    def table_ddl(self):
        return self.__table_ddl

    @property
    def table_path(self):
        return self.__table_path

    @table_path.setter
    def table_path(self, table_path: str):
        self.__table_path = table_path

    @property
    def table_schema(self):
        return self.__table_schema

    @table_schema.setter
    def table_schema(self, table_schema: str):
        self.__table_schema = table_schema


class OutputTable(SimioTable):
    def __init__(self, model_metadata, table_path: str):
        """
        Child class of SimioTable for the output tables.
        :param model_metadata:
        :param table_path: path where the tables lay.
        """
        SimioTable.__init__(self,
                            model_metadata,
                            table_path)

    @staticmethod
    def get_single_row_value(dom_row_list: list, attribute_name='Value'):
        """
        Get value of a single scenario's response.
        :param dom_row_list: Single row provided as a list of one term.
        :param attribute_name: attribute reporting the value of the cell.
        :return: Value of such observation.
        """
        row_list = extract_list_from_dom(dom_object=dom_row_list,
                                         attribute_name=attribute_name)
        if len(row_list) == 0:
            response_value = np.NaN
        else:
            response_value = row_list[0]
        return response_value


class InputTable(SimioTable):
    def __init__(self, model_metadata, table_path: str):
        """
        Child class of SimioTable for the input tables.
        :param model_metadata:
        :param table_path: path where the tables lay.
        """
        SimioTable.__init__(self,
                            model_metadata,
                            table_path)

    @staticmethod
    def get_single_row_value(dom_row_list: list, tag_value='Value'):
        """
        Get value of a single scenario's response. If the value is not nan, when there is a value with tag 'Value'
        report that, otherwise report the childNodes of the 'Property' Tag.
        :param dom_row_list: Single row provided as a list of one term.
        :param tag_value: name of the tag reporting the value of the cell
        :return: Value of such observation.
        """
        response = dom_row_list[0]
        if pd.isna(response):
            response_value = np.NaN
        elif len(response.getElementsByTagName(tag_value)) > 0:
            response_value = response.getElementsByTagName(tag_value)[0].childNodes[0].data
        else:
            response_value = response.childNodes[0].data
        return response_value


class SimioTables:
    def __init__(self,
                 path_to_project: str,
                 model_file_name: str,
                 model_name: str):
        self.__model_metadata = get_model_metadata(path_to_project=path_to_project,
                                                   model_file_name=model_file_name,
                                                   model_name=model_name)
        self.__tables_path = get_model_tables_path(project_path=path_to_project,
                                                   project_filename=model_file_name,
                                                   model_name=model_name)
        self.__output_file_list = get_output_table_names(output_table_path=self.__tables_path)
        self.__input_file_list = get_input_table_names(input_table_path=self.__tables_path)
        self.__output_tables = None
        self.__input_tables = None
        self.__input_tables_ddl = None
        self.__output_tables_ddl = None

    def load_output_tables(self):
        """
            Load all experiment results related to a Simio model.
            :return: Dictionary whose keys are experiment name and value is a data frame (or None).
            """
        experiment_dictionary = dict()
        schema_list = list()
        for _table_name in self.output_file_list:
            table_path = os.path.join(self.tables_path, _table_name)
            output_table = OutputTable(model_metadata=self.model_metadata,
                                       table_path=table_path)
            experiment_dictionary[_table_name.split('.')[0]] = output_table.load_single_table(tag_name='StateValue',
                                                                                              attribute_name='Name')
            schema_list.append(output_table.table_ddl)
        self.output_tables = experiment_dictionary
        self.output_tables_ddl = '\n'.join(schema_list)

    def load_input_tables(self):
        """
            Load all experiment results related to a Simio model.
            :return: Dictionary whose keys are experiment name and value is a data frame (or None).
            """
        experiment_dictionary = dict()
        schema_list = list()
        for _table_name in self.input_file_list:
            table_path = os.path.join(self.tables_path, _table_name)
            input_table = InputTable(model_metadata=self.model_metadata,
                                     table_path=table_path)
            experiment_dictionary[_table_name.split('.')[0]] = input_table.load_single_table(tag_name='Property',
                                                                                             attribute_name='Name')
            schema_list.append(input_table.table_ddl)
        self.input_tables = experiment_dictionary
        self.input_tables_ddl = '\n'.join(schema_list)

    # Getters and Setters
    @property
    def input_tables_ddl(self):
        return self.__input_tables_ddl

    @input_tables_ddl.setter
    def input_tables_ddl(self,
                         new_tables_schema):
        self.__input_tables_ddl = new_tables_schema

    @property
    def output_tables_ddl(self):
        return self.__output_tables_ddl

    @output_tables_ddl.setter
    def output_tables_ddl(self,
                          new_tables_schema):
        self.__output_tables_ddl = new_tables_schema

    @property
    def model_metadata(self):
        return self.__model_metadata

    @model_metadata.setter
    def model_metadata(self,
                       model_metadata_xml):
        self.__model_metadata = model_metadata_xml

    @property
    def tables_path(self):
        return self.__tables_path

    @tables_path.setter
    def tables_path(self,
                    path_to_tables: str):
        self.__tables_path = path_to_tables

    @property
    def output_file_list(self):
        return self.__output_file_list

    @output_file_list.setter
    def output_file_list(self,
                         tables_path: str):
        self.__output_file_list = get_output_table_names(output_table_path=tables_path)

    @property
    def input_file_list(self):
        return self.__input_file_list

    @input_file_list.setter
    def input_file_list(self,
                        tables_path: str):
        self.__input_file_list = get_input_table_names(input_table_path=tables_path)

    @property
    def input_tables(self):
        return self.__input_tables

    @input_tables.setter
    def input_tables(self,
                     input_tables: dict):
        self.__input_tables = input_tables

    @property
    def output_tables(self):
        return self.__output_tables

    @output_tables.setter
    def output_tables(self,
                      output_tables: dict):
        self.__output_tables = output_tables


if __name__ == '__main__':
    pass
